﻿using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Xml;
using System.Xml.Schema;
using Alternet.Common;
using Alternet.Syntax;

namespace XMLJSONValidator
{
    public class ValidatingXmlParser : XmlParser
    {
        private XmlReaderSettings settings;
        private XmlSchema schema;
        private IList<SyntaxError> semanticErrors = new List<SyntaxError>();

        public ValidatingXmlParser()
        {
            settings = new XmlReaderSettings();
            settings.ValidationType = ValidationType.Schema;
            settings.ValidationFlags = XmlSchemaValidationFlags.ReportValidationWarnings;
            settings.ValidationEventHandler += ValidationHandler;
        }

        public XmlSchema Schema
        {
            get
            {
                return schema;
            }

            set
            {
                if (schema != value)
                {
                    schema = value;
                    settings.Schemas = new XmlSchemaSet();

                    if (schema != null)
                        settings.Schemas.Add(schema);
                }
            }
        }

        private IRange FindRange(string line, Point point)
        {
            int start = point.X;
            while (start > 0 && char.IsLetterOrDigit(line[start - 1]))
                start--;

            int end = point.X;
            while (end < line.Length && char.IsLetterOrDigit(line[end]))
                end++;

            return new Range(new Point(start, point.Y), new Point(end, point.Y));
        }

        public void ValidationHandler(object sender, ValidationEventArgs e)
        {
            var lineIndex = e.Exception.LineNumber - 1;
            if (lineIndex < 0 || lineIndex >= Strings.Count)
                return;

            var line = Strings[lineIndex];
            var characterIndex = MathUtilities.Clamp(e.Exception.LinePosition - 1, 0, line.Length - 1);

            var error = new SyntaxError(new Point(characterIndex, lineIndex), string.Empty, e.Message);

            error.ErrorType = e.Severity == XmlSeverityType.Error ? SyntaxErrorType.Warning : SyntaxErrorType.Info;
            error.Range = FindRange(line, error.Position);

            semanticErrors.Add(error);
        }

        public override int GetSyntaxErrors(IList<ISyntaxError> errors)
        {
            base.GetSyntaxErrors(errors);

            foreach (var err in semanticErrors)
                errors.Add(err);
            return errors.Count;
        }

        public override void ReparseText()
        {
            base.ReparseText();

            semanticErrors.Clear();

            if (schema == null)
                return;

            using (var stringReader = new StringReader(Strings.Text))
            {
                var xmlReader = XmlReader.Create(stringReader, settings);
                var document = new XmlDocument();
                try
                {
                    document.Load(xmlReader);
                    document.Validate(ValidationHandler);
                }
                catch
                {
                }
            }
        }

        protected override void AssignParser(ISyntaxParser parser)
        {
            base.AssignParser(parser);
            ((ValidatingXmlParser)parser).Schema = Schema;
        }
    }
}